/* dump entry for extract */
#define N_(x) x

/* remember to update "default" in the same directory. */

const char* dump = {
    N_("Pinyin Symbol"),
    N_("Chewing Symbol"),
    N_("Cangjie"),
    N_("Pinyin"),
    N_("Tone Pinyin"),
    N_("Keyboard"),
    N_("Anthy"),
    N_("Hanja"),
    N_("Romaja"),
    N_("Quick")
};
